// GSMlib stuff

#include "gsm.h"

#include <string>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <dirent.h>
#include <signal.h>
#include <fstream>
#include <iostream>

#include <gsmlib/gsm_me_ta.h>
#include <gsmlib/gsm_event.h>
#include <gsmlib/gsm_unix_serial.h>

#include <glibmm.h>

#include <time.h>


//  add_port() (port, baud, initstring)
//  remove_port()

// PhoneListener class

PhoneListener::PhoneListener()
{
m_terminateSent=false;
  m_erase_mode=false;
}

void
PhoneListener::stop()
{
  Glib::Mutex::Lock lock(mutex);
  m_terminateSent=true;
  lock.release();

}

bool
PhoneListener::terminateSent()
{
  bool status=false;
  Glib::Mutex::Lock lock(mutex);
  status=m_terminateSent;
  lock.release();
  return status;
}

void
PhoneListener::queue_outgoing_message(Glib::ustring num,
									  Glib::ustring msg)
{
  Glib::Mutex::Lock lock(mutex);
  OutgoingMessage m;
  m.number=num;
  m.message=msg;
  sendQueue.push_back(m);
  lock.release();
  g_message("Queued message to %s", num.c_str());
}

void
PhoneListener::send_message(gsmlib::MeTa *mt, OutgoingMessage *m)
{
  gsmlib::Ref<gsmlib::SMSSubmitMessage> submitSMS =
	new gsmlib::SMSSubmitMessage();
  gsmlib::Address destAddr(m->number);
  submitSMS->setDestinationAddress(destAddr);
  g_message("Sending message to %s", m->number.c_str());
  mt->sendSMSs(submitSMS, m->message, true);
  g_message("Message sent");
}

PhoneListener::notify_sig_t
PhoneListener::signal_notify()
{
  return m_signal_notify;
}

PhoneListener::status_sig_t
PhoneListener::signal_status()
{
  return m_signal_status;
}

void
PhoneListener::send_message_notification(gsmlib::SMSMessageRef msg)
{
  struct tm stamp;
  gsmlib::Address addr=msg->address();
  GTime timestamp;
  Glib::ustring datestring;
  Glib::Date *date;
  gchar buf[32];
  int y;

  // TODO: convert this code to use ICU, which does the
  // job properly

  stamp.tm_sec=msg->serviceCentreTimestamp()._seconds;
  stamp.tm_min=msg->serviceCentreTimestamp()._minute;
  stamp.tm_hour=msg->serviceCentreTimestamp()._hour;
  stamp.tm_mday=msg->serviceCentreTimestamp()._day;
  stamp.tm_mon=msg->serviceCentreTimestamp()._month -1;
  y=msg->serviceCentreTimestamp()._year;
  stamp.tm_year=y < 80 ? y+100 : y;
  stamp.tm_isdst = -1;
  stamp.tm_yday=0;
  stamp.tm_wday=0;

  timestamp=(GTime)mktime(&stamp);

  date=new Glib::Date();
  date->set_time(timestamp);
  datestring=date->format_string("%x");
  delete date;

  datestring+=" ";
  strftime(buf, sizeof(buf), "%X", &stamp);
  datestring+=buf;

  m_signal_notify.emit(addr.toString(),
					   datestring,
					   msg->userData());
}

void
PhoneListener::sms_loop(std::string device, Glib::Dispatcher *disp)
{
  m_signal_status.emit(PHONELISTENER_CONNECTING);

  try {
	// open GSM device
	gsmlib::UnixSerialPort *port=new gsmlib::UnixSerialPort(device,
									gsmlib::DEFAULT_BAUD_RATE,
									gsmlib::DEFAULT_INIT_STRING,
									false);

	// note the MeTa deletes the port in its deconstructor.

	// set 5 second timeout on serial port
	port->setTimeOut(5);
	gsmlib::MeTa m(port);

	std::string dummy1, dummy2, receiveStoreName;
	m.getSMSStore(dummy1, dummy2, receiveStoreName);
	m.setMessageService(1);

	// switch on SMS routing
	m.setSMSRoutingToTA(true, false, false,
						true); // only Receiption indication

	m_signal_status.emit(PHONELISTENER_CONNECTED);

	// register event handler
	EventHandler *e=new EventHandler();
	e->setStoreName(receiveStoreName);
	e->setMessageQueue(&newMessages);
	m.setEventHandler(e);

	// wait for new messages
	bool exitScheduled = false;
	while (1) {
	  struct timeval timeoutVal;
	  timeoutVal.tv_sec = 1;
	  timeoutVal.tv_usec = 0;
	  m.waitEvent(&timeoutVal);

	  // if it returns, there was an event or a timeout
	  while (newMessages.size() > 0) {
	    // get first new message and remove it from the vector
	    gsmlib::SMSMessageRef newSMSMessage = newMessages.begin()->_newSMSMessage;
	    gsmlib::CBMessageRef newCBMessage = newMessages.begin()->_newCBMessage;
	    gsmlib::GsmEvent::SMSMessageType messageType =
	      newMessages.begin()->_messageType;
	    int index = newMessages.begin()->_index;
	    std::string storeName = newMessages.begin()->_storeName;
	    newMessages.erase(newMessages.begin());

	    // process the new message
	    std::string result = "Type of message: ";
	    switch (messageType)
	    {
	    case gsmlib::GsmEvent::NormalSMS:
	      result += "SMS message\n";
	      break;
	    default:
	      result += "unknown message\n";
	      break;
	    }

	    if (!newSMSMessage.isnull())
			{
	      result += newSMSMessage->toString();
			  send_message_notification(newSMSMessage);
			  g_message("SMS Message");
			}
	    else if (! newCBMessage.isnull())
			{
	      result += newCBMessage->toString();
			  g_message("CBMessage");
			}
	    else
	    {
	      gsmlib::SMSStoreRef store = m.getSMSStore(storeName);

	      if (messageType == gsmlib::GsmEvent::CellBroadcastSMS) {
				g_message("Broadcast message received");
		result += (*store.getptr())[index].cbMessage()->toString();
			  }
	      else
			  {
				g_message("SMS message received");
		result += (*store.getptr())[index].message()->toString();
				send_message_notification((*store.getptr())[index].message());
	      }

	      if (m_erase_mode) {
				g_message("Erasing message from phone");
				store->erase(store->begin() + index);
			  }
	    }
	    g_message(result.c_str());
	  }

	  // if no new SMS came in and program exit was scheduled, then exit
	  if (exitScheduled) {
		m_signal_status.emit(PHONELISTENER_IDLE);
		return;
	  }

	  // handle terminate signal
	  if (terminateSent()) {
	    exitScheduled = true;
		m_signal_status.emit(PHONELISTENER_DISCONNECTING);
	    // switch off SMS routing
	    try {
		  m.setSMSRoutingToTA(false, false, false);
		} catch (gsmlib::GsmException &ge) {
	      // some phones (e.g. Motorola Timeport 260) don't allow to switch
	      // off SMS routing which results in an error. Just ignore this.
	    }
	    // the AT sequences involved in switching of SMS routing
	    // may yield more SMS events, so go round the loop one more time
	  }

	  // see if there's anything to send, if so send one

	  if (sendQueue.size()>0) {
		  Glib::Mutex::Lock lock(mutex);
		  send_message(&m, &(*sendQueue.begin()));
		  sendQueue.erase(sendQueue.begin());
		  lock.release();
	  }

	} // while (1)

  } catch (gsmlib::GsmException &ge) {
	cerr << ("[ERROR]: ") << ge.what() << endl;
	if (ge.getErrorClass() == gsmlib::MeTaCapabilityError)
	  cerr << ("[ERROR]: ")
		   << ("(try setting sms_type, please refer to gsmsmsd manpage)")
		   << endl;

	m_signal_status.emit(PHONELISTENER_ERROR);

	// emit dispatched signal to signify exit by error
	(*disp)();
	return;
  }

  // control never reaches here, how pants
  return;
}

void
PhoneListener::init()
{
  // do nothing for now
}

Glib::Thread *
PhoneListener::sms_loop_thread(std::string device, Glib::Dispatcher *disp)
{
  return Glib::Thread::create(SigC::bind<string,Glib::Dispatcher*>
							  (SigC::slot(*this, &PhoneListener::sms_loop),
							   device, disp),
							  true);
}

// EventHandler class

void
EventHandler::setStoreName(std::string rs)
{
  receiveStoreName=rs;
}

void
EventHandler::setMessageQueue(vector<IncomingMessage> *nm)
{
  newMessages=nm;
}

void EventHandler::SMSReception(gsmlib::SMSMessageRef newMessage,
				gsmlib::GsmEvent::SMSMessageType messageType)
{
  IncomingMessage m;
  m._messageType = messageType;
  m._newSMSMessage = newMessage;
  newMessages->push_back(m);
}

void EventHandler::CBReception(gsmlib::CBMessageRef newMessage)
{
  IncomingMessage m;
  m._messageType = gsmlib::GsmEvent::CellBroadcastSMS;
  m._newCBMessage = newMessage;
  newMessages->push_back(m);
}

void EventHandler::SMSReceptionIndication(std::string storeName, 
										  unsigned int index,
							  gsmlib::GsmEvent::SMSMessageType messageType)
{
  IncomingMessage m;
  m._index = index;

  if (receiveStoreName != "" && ( storeName == "MT" || storeName == "mt"))
    m._storeName = receiveStoreName;
  else
    m._storeName = storeName;

  m._messageType = messageType;
  newMessages->push_back(m);
}
